<?php
namespace Importacao\Logic\Bases\ModeloImportacao;

use \Importacao\Logic\Bases\ModeloImportacao\EntityContainer;
use \Importacao\Logic\Bases\ModeloImportacao\BasicEntity;
use \Importacao\Logic\Bases\ModeloImportacao\References\Autoincrement;
use \Importacao\Logic\Bases\ModeloImportacao\References\Executavel;
use \Importacao\Logic\Bases\ModeloImportacao\References\FixedValue;
use \Importacao\Logic\Bases\ModeloImportacao\References\ForeignKey;
use \Importacao\Logic\Bases\ModeloImportacao\References\Label;
use \Importacao\Logic\Bases\ModeloImportacao\References\PadraoInvalido;
use \Importacao\Logic\Bases\ModeloImportacao\References\Reference;
use \Importacao\Logic\Bases\ModeloImportacao\References\ValorCampo;
use \Importacao\Logic\Bases\ModeloImportacao\References\FixedField;
use \Importacao\Logic\Integracao\ProjetoImportacao;
use \Importacao\Logic\Util\Util;

/**
 * Constroi o conjunto de dados que sera utilizado para a criacao das queries da importacao.
 *
 * Class DataBuilder
 * @package Importacao\Logic\Bases\ModeloImportacao
 */
class DataBuilder
{
    const PADRAO_SET     = 'set';
    const CDOPERINCLUSAO = '000000099991';

    private $agrupamento;
    private $data;
    private $agrupadores;
    private $rules;
    private $fixedFields;
    private $padroesMetadados;
    private $entidades;
    private $nrorg;
    private $dataAtual;
    private static $existeErro = false;
    private static $countError = 0;

    static $pk = array();

    /**
     *
     *
     * @param $agrupamento array Relacao entre os dados dos arquivos excel e as tabelas do banco de dados (esturura.json).
     * @param $data        array Array com todos os dados lidos dos arquivos de excel.
     * @param $helpers     array Array com todos os dados auxiliares do modelo de importacao (helper.json).
     */
    public function __construct($agrupamento, $data, $helpers, $nrorg = null)
    {
        $this->dataAtual       = date('d/m/Y');
        $this->nrorg           = $nrorg;
        $this->agrupamento     = $agrupamento;
        $this->data            = $data;
        $this->rules           = $helpers["rules"];
        $this->fixedFields     = $helpers["fixedFields"];
        // Armazena o nome dos agrupadores (presentes no arquivo estrutura.json) que serão utilizados em determinado projeto de importacao.
        $this->agrupadores     = array();
        // Objeto onde todas as entidades criadas serao planificadas. Este array sera retornado no final do processo de geracao dos dados.
        // É publicamente acessivel pelo metodo addAgrupador.
        $this->entidades       = array();

        // Este array armazena todos os padroes que sao compreendidos pelo modelo, e sao utilizados nos arquivos de estrutura.json e helper.json. Qualquer classe
        // utilizada neste array deverá implementar a interface de \Importacao\Logic\Bases\ModeloImportacao\References\Reference.
        $this->padroesMetadados = array(
            new ValorCampo(),
            new ForeignKey(),
            new Label(),
            new FixedField(),
            new FixedValue(),
            new Executavel(),
            new PadraoInvalido(),
        );
    }

    public function addAgrupador($agrupador)
    {
        $this->agrupadores[] = $agrupador;
    }

    public function build()
    {
        $this->percorreLinhasDados();
        // Retorna todos os dados planos e insere todos os campos padrao.
        return $this->entidades;
    }

    /**
     * Percorre todas as linhas do excel.
     */
    private function percorreLinhasDados()
    {
        $progressBar = new \Importacao\Logic\Util\BarraProgressoCli(
            'Criando Arquivo SQL', count($this->data)
        );

        foreach ($this->data as $key => $linhaDados) {
            $this->percorreAgrupamentos($linhaDados);
            $status = $progressBar->showStatus($key + 1);
            if (!empty($status)) {
                $dir_fileLog = __DIR__."/../../../../../../../../mobile/files/uploads/".$this->nrorg."/results/";
                file_put_contents(realpath($dir_fileLog).'\status.log', $status.PHP_EOL, FILE_APPEND);
            }
            unset($this->data[$key]);
        }

        $progressBar->finish();
    }

    /**
     * Percorre todos os agrupamentos do arquivo estrutura.json
     *
     * @param $linhaDados array objeto contendo a linha corrente do excel. Passa por este metodo somente para ser transmitido para os proximos,
     * onde este realmente sera necessario.
     */
    public function percorreAgrupamentos($linhaDados)
    {
        foreach ($this->agrupamento as $nomeAgrupamento => $tabelas) {
            if (in_array($nomeAgrupamento, $this->agrupadores)) {
                /* Exibe log do tipo de importação do registro na execução do consumidor */
                echo "-- Tipo de Importacao: ". $nomeAgrupamento.PHP_EOL;

                $entityContainer = new EntityContainer();
                $this->percorreTabelas($linhaDados, $tabelas, $entityContainer, $nomeAgrupamento);
//                $this->impedeRepetirPk($entityContainer);
            }
        }
    }

    /**
     * Percorre todas as regras descritas para determinada tabela (se estas existirem). Apos executar todas as regras desta tabela,
     * este metodo tambem faz chamada ao metodo que executa as regras para os campos destas tabelas.
     *
     * As regras sao descritas no arquivo helper.json, sendo que cada regra deve conter as chaves de "condicao", onde é descrita
     * a condicao a ser verificada, "entao", onde e descrito o que deve ser feito, caso a condicao seja satisfeita, e "senao", que
     * descreve o que deve ser feito caso a condicao nao seja satisfeita.
     *
     * Importante notar que, sempre que a condicao resultar em "false", a entidade (tabela) em questao sera removida do result.
     *
     * @param $regras                          array Array contendo o conjunto condica, entao, senao
     * @param EntityContainer $entityContainer       Container armazenando todas as entidades de determinada linha.
     * @param $linhaDados                      array Array contendo os dados da linha do arquivo excel.
     */
    private function executaRegrasDeTabelas($regras, EntityContainer $entityContainer, $linhaDados)
    {
        foreach ($regras["tabelas"] as $nomeTabela => $tabela) {
            foreach ($tabela["regras"] as $regra) {
                $condicao = $this->verificaCondicaoRegra(
                    $regra["condicao"], $nomeTabela, $entityContainer, $linhaDados
                );
                if ($condicao) {
                    $this->getDadosReference(
                        $nomeTabela, null, $regra["entao"], null, $entityContainer, $linhaDados
                    );
                } else {
                    $entityContainer->removeTabela($nomeTabela);
                    if (isset($regra["senao"])) {
                        $this->getDadosReference(
                            $nomeTabela, null, $regra["senao"], null, $entityContainer, $linhaDados
                        );
                    }
                }
            }

            if (isset($tabela["colunas"])) {
                $this->executaRegrasDeColuna($tabela["colunas"], $nomeTabela, $entityContainer, $linhaDados);
            }
        }
    }

    /**
     * Executa as regras a nivel de coluna. Assim como nas regras de tabela, estas devem possuir condicao, entao e senao.
     *
     * @param $regrasColuna
     * @param $nomeTabela
     * @param EntityContainer $entityContainer
     * @param $linhaDados
     * @return bool
     */
    private function executaRegrasDeColuna($regrasColuna, $nomeTabela, EntityContainer $entityContainer, $linhaDados)
    {
        /** @var  $entidade BasicEntity*/
        $entidade = $entityContainer->getTabela($nomeTabela);
        if (!$entidade) {
            return false;
        }

        foreach ($regrasColuna as $colunaName => $coluna) {
            foreach ($coluna['regras'] as $regra) {
                $condicao = $this->verificaCondicaoRegra(
                    $regra["condicao"], $nomeTabela, $entityContainer, $linhaDados
                );

                if ($condicao) {
                    $result = $this->getDadosReference(
                        $nomeTabela, null, $regra["entao"], null, $entityContainer, $linhaDados
                    );

                    $entidade->addField($colunaName, $result, $entidade->getTableName());
                } else if (isset($regra["senao"])) {
                    $result = $this->getDadosReference(
                        $nomeTabela, null, $regra["senao"], null, $entityContainer, $linhaDados
                    );
                    $entidade->addField($colunaName, $result, $entidade->getTableName());
                }
            }
        }
        return true;
    }

    /**
     * Metodo utilizado para resolucao da condicao.
     *
     * @param $regra          string            Padrao informado no helper.json, que apos processado, possui a condicao necessaria.
     * @param $nomeTabela     string            Nome da tabela sobre a qual a condicao sera imposta
     * @param EntityContainer $entityContainer  Container com todas as entidades de determinada linha.
     * @param $linhaDados     array             Linha de dados do excel
     *
     * @return boolean
     */
    private function verificaCondicaoRegra($regra, $nomeTabela, EntityContainer $entityContainer, $linhaDados)
    {
        $condicao = $this->getDadosReference($nomeTabela, null, $regra, null, $entityContainer, $linhaDados);
        $result = null;
        eval('$result = ('.$condicao.');');
        return $result;
    }

    /**
     * Percorre todas as tabelas de determinado agrupamento do arquivo de estruturas (estrutura.json).
     *
     * @param $linhaDados     array             Linha do excel. Nao e utilizado por este metodo, somente transmitido para os
     *                                          metodos posteriores.
     * @param $tabelas        array             Array com todas as tabelas do agrupamento corrente.
     * @param EntityContainer $entityContainer  Container que devera armazenar todas as entidades geradas a partir da linha.
     * @param $nomeAgrupamento
     */
    private function percorreTabelas($linhaDados, $tabelas, EntityContainer $entityContainer, $nomeAgrupamento) {
        foreach ($tabelas as $nomeTabela => $dadosTabela) {
            self::$countError = 0;            

            \Setforms2\Shared\Registry::getEntityManager()->beginTransaction();

            $entityContainer->adicionaTabela($nomeTabela, new BasicEntity($nomeTabela));
            $this->percorreDados($linhaDados, $nomeTabela, $dadosTabela, $entityContainer);
            $this->insereCamposPadrao($entityContainer->getTabela($nomeTabela));

            if (isset($this->rules[$nomeAgrupamento])) {
                $this->executaRegrasDeTabelas($this->rules[$nomeAgrupamento], $entityContainer, $linhaDados);
            }

            $this->impedeRepetirPk($entityContainer);

            // persiste os dados na entidade, se não existir erro, persite o registro na entidade
            if (self::$countError == 0 && !empty($entityContainer->getTabela($nomeTabela))) {
                $nomeTabela = \Importacao\Logic\Util\Util::normalizeNameClass($nomeTabela);
                CarregaEntidade::getInstance()->getEntidade($nomeTabela)->persist();
                \Setforms2\Shared\Registry::getEntityManager()->flush();
                \Setforms2\Shared\Registry::getEntityManager()->commit();
                //\Setforms2\Shared\Registry::getEntityManager()->clear();
                CarregaEntidade::getInstance()->inicializeEntidade($nomeTabela);

                /* Exibe log de sucesso na inserção do registro na execução do consumidor */
                echo "-- Inserido com sucesso na tabela: ". $nomeTabela.PHP_EOL;
            } else {
                /* Exibe log de erro na inserção do registro na execução do consumidor */
                echo "-- Erro ao inserir na tabela: ". $nomeTabela.PHP_EOL;
            }
        }

        // atualiza se a validação foi executada COM erros ou SEM erros.
        $this->validacaoStatus(self::$existeErro);
    }

    /**
     * Percorre todos os dados de uma linha do excel.
     *
     * @param $linhaDados     array            Linha do excel.
     * @param $nomeTabela     string           Nome da tabela.
     * @param $dadosTabela    array            conteudo das tabelas descritas em estrutura.json. Nao e utilizado aqui, mas
     *                                         transmitido para os metodos posteriores.
     * @param EntityContainer $entityContainer Container que devera armazenar todas as entidades geradas a partir da linha.
     */
    private function percorreDados($linhaDados, $nomeTabela, $dadosTabela, EntityContainer $entityContainer)
    {
        $this->addPkTabela($nomeTabela, $dadosTabela, $entityContainer);
        foreach ($linhaDados as $nomeColunaDados => $valorColunaDados) {
            $this->percorreDadosAgrupamento (
                $nomeTabela, $dadosTabela, $nomeColunaDados, $valorColunaDados, $linhaDados, $entityContainer
            );
        }
        $this->addPkValor($nomeTabela, $entityContainer);
    }

    /**
     * Percorre todos os dados de uma tabela, dentro de um agrupamento.
     *
     * @param $nomeTabela       string
     * @param $dadosTabela
     * @param $nomeColunaDados
     * @param $valorColunaDados
     * @param $linhaDados
     * @param EntityContainer $entityContainer
     */
    private function percorreDadosAgrupamento($nomeTabela, $dadosTabela, $nomeColunaDados,
                                              $valorColunaDados, $linhaDados, EntityContainer $entityContainer)
    {
        foreach ($dadosTabela as $nomeColuna => $dadosColuna) {
            $this->constroiFieldEntidade(
                $nomeTabela, $nomeColunaDados, $valorColunaDados, $nomeColuna,
                $dadosColuna, $linhaDados, $entityContainer
            );
        }
    }

    private function constroiFieldEntidade($nomeTabela, $nomeColunaDados, $valorColunaDados,
                                           $nomeColuna, $dadosColuna, $linhaDados, EntityContainer $entityContainer)
    {
        /** @var  $entidade BasicEntity*/
        $entidade        = $entityContainer->getTabela($nomeTabela);

        if ($nomeColuna === $nomeColunaDados) {
            $this->addFields($entidade, $dadosColuna, $valorColunaDados);
        } else if (preg_match("/\#/", $nomeColuna) && !$entidade->hasColumn($nomeColuna)) {
            $this->adicionaCamposReference(
                $nomeTabela, $nomeColuna, $dadosColuna, $valorColunaDados, $entidade, $linhaDados, $entityContainer
            );
        }
    }

    private function adicionaCamposReference($nomeTabela, $nomeColuna, $dadosColuna, $valorColunaDados,
                                             BasicEntity $entidade, $linhaDados, EntityContainer $entityContainer)
    {
        $dadoColuna = $this->getDadosReference (
            $nomeTabela, $nomeColuna, $dadosColuna, $valorColunaDados, $entityContainer, $linhaDados
        );

        if (!preg_match("/".Reference::PADRAO_INVALIDO."/", $dadoColuna)) {
            $entidade->addField(str_replace("#", "", $nomeColuna), $dadoColuna, $entidade->getTableName());
        }
    }

    private function addFields(BasicEntity $entidade, $dadosColuna, $valorColunaDados)
    {
        $fields = explode(",", str_replace(" ", "", $dadosColuna));
        foreach ($fields as $field) {
            $entidade->addField($field, $valorColunaDados, $entidade->getTableName());
        }
    }

    // Retorna o valor esperado para uma coluna informada nos agrupamentos através de meta-dados.
    private function getDadosReference($nomeTabela, $nomeColuna, $metaDado, $valorFixo,
                                       EntityContainer $entityContainer, $linhaDados = null)
    {
        $padraoEncontrado = false;
        $matches = array();
        /** @var $padrao \Importacao\Logic\Bases\ModeloImportacao\References\Reference */
        foreach ($this->padroesMetadados as $padrao) {
            if (preg_match_all($padrao->getRegularExpression(), $metaDado, $matches)) {
                /* Itera a ocorrencia de varios padroes juntos */
                for ($i = count($matches[0]); $i > 0; $i--) {
                    $metaDado = $padrao->getValorReferencia(
                        $nomeTabela, $nomeColuna, $metaDado, $valorFixo, $entityContainer, $linhaDados, $this->fixedFields
                    );
                }
                $padraoEncontrado = true;
            }
        }

        return !$padraoEncontrado ? $metaDado : $this->getDadosReference (
            $nomeTabela, $nomeColuna, $metaDado, $valorFixo, $entityContainer, $linhaDados
        );
    }

    private function insereCamposPadrao(BasicEntity $entidade)
    {
        foreach ($this->fixedFields as $name => $value) {
            if (!$entidade->hasColumn($name)) {
                if ($entidade->getTableName() != 'MIGRAMDEPARA') {
                    //$entidade->addField($name, $value, $entidade->getTableName());
                    if ($name == 'NRORG') {
                        $entidade->addField($name, $this->nrorg, $entidade->getTableName());                                            
                    } elseif ($name == 'DTINCLUSAO') {
                        $entidade->addField($name, $this->dataAtual, $entidade->getTableName());
                    } elseif ($name == 'DTULTATU') {
                        $entidade->addField($name, $this->dataAtual, $entidade->getTableName());
                    } elseif ($name == 'NRORGINCLUSAO') {
                        $entidade->addField($name, $this->nrorg, $entidade->getTableName());
                    } elseif ($name == 'CDOPERINCLUSAO') {
                        $entidade->addField($name, self::CDOPERINCLUSAO, $entidade->getTableName());
                    } elseif ($name == 'NRORGULTATU') {
                        $entidade->addField($name, $this->nrorg, $entidade->getTableName());
                    }
                } else {
                    if ($name == 'NRORG') {
                        $entidade->addField($name, $this->nrorg, $entidade->getTableName());                                            
                    }
                }
            }
        }
    }

    /**
     * @param $nomeTabela
     * @param $dadosTabela
     * @param EntityContainer $entityContainer
     */
    private function addPkTabela($nomeTabela, $dadosTabela, EntityContainer $entityContainer)
    {
        if (isset($dadosTabela['PK'])) {
            $entityContainer->getTabela($nomeTabela)->setPk($dadosTabela['PK']);
            unset($dadosTabela['PK']);
        }
    }

    /**
     * @param $nomeTabela
     * @param EntityContainer $entityContainer
     */
    private function addPkValor($nomeTabela, EntityContainer $entityContainer)
    {
        $pk = $entityContainer->getTabela($nomeTabela)->getPk();
        if (!$pk) return;
        foreach ($pk as $key => $value) {
            $pk[$key] = $entityContainer->getTabela($nomeTabela)->getField($key);
        }
        $entityContainer->getTabela($nomeTabela)->setPk($pk);
    }

    /**
     * @param EntityContainer $entityContainer
     */
    private function impedeRepetirPk(EntityContainer $entityContainer)
    {
        foreach ($entityContainer->getTabelas() as $tabela) {
            $repete = false;
            if ($tabela->getPk()) {
                $pks = isset(self::$pk[$tabela->getTableName()]) ? self::$pk[$tabela->getTableName()] : null;
                $isset = isset(self::$pk[$tabela->getTableName()]);
                if (isset(self::$pk[$tabela->getTableName()])) {
                    foreach (self::$pk[$tabela->getTableName()] as $pk) {
                        if ($pk && ($pk == $tabela->getPk())) {
                            $repete = true;
                            break;
                        }
                    }
                }
                if (!$repete) {
                    self::$pk[$tabela->getTableName()][] = $tabela->getPk();
                }
            }
            if (!$repete) {
                array_push($this->entidades, $tabela);
            }
//            if (!$repete) {
//                 array_push($this->entidades, $tabela);
//            } else if ($tabela->getTableName() != "FPA_CALCULOFOLHA") {
//                // \Importacao\Logic\Integracao\ProjetoImportacao::getInstancia()->criaLogDeItemProjeto(null, "Registro Duplicado na tabela {$tabela->getTableName()}", '2', null, \Geral\Entities\Logprojimport::VALIDACAO_INTEGRIDADE);
//            }
        }
    }

    /**
     * @param $entidade
     * @param $fieldName
     * @param $fieldValue
     * @return bool
     * @internal param $field
     * @internal param $valorColunaDados
     * @internal param $tablename
     */
    public static function carregaEntidade($entidade, $fieldName, $fieldValue)
    {
        $tablename = \Importacao\Logic\Util\Util::normalizeNameClass($entidade);
        $fieldName = self::PADRAO_SET.ucfirst(strtolower($fieldName));

        try {
//            echo "$tablename - $fieldName - $fieldValue".PHP_EOL;
            CarregaEntidade::getInstance()->getEntidade($tablename)->$fieldName(trim($fieldValue));
            CarregaEntidade::getInstance()->getEntidade($tablename)->setNrprojimport(ProjetoImportacao::getNrprojetoimport());
        } catch (\Exception $e) {
            $msg = $e->getMessage();
            if (!empty($msg)) {
                \Importacao\Logic\Integracao\ProjetoImportacao::getInstancia()->criaLogDeItemProjeto(null, $e->getMessage(), null, null, \Geral\Entities\Logprojimport::VALIDACAO_INTEGRIDADE);
                self::$existeErro = true;
                self::$countError++;
            }
        }
    }

    /**
     * @param $fieldName
     * @return string
     */
    public function normaliseNameField($fieldName)
    {
        return self::PADRAO_SET.ucfirst(strtolower($fieldName));
    }

    /**
     * @param $existeErro
     */
    public function validacaoStatus($existeErro)
    {
        if ($existeErro) {
            \Importacao\Logic\Integracao\ProjetoImportacao::getInstancia()->atualizaPercentStatus("Validação com erros");
        } else {
            \Importacao\Logic\Integracao\ProjetoImportacao::getInstancia()->atualizaPercentStatus("Validação sem erros");
            // atualiza status para concluído
            ProjetoImportacao::getInstancia()->atualizaItemArqImportado(\Geral\Entities\Itemprojimport::STATUS_CONCLUIDO);
        }
    }

}